from wq.io.mappers import TupleMapper
from wq.io.gis.mixins import GisParser, GisLoader, WktMapper
from wq.io.base import BaseIO


class GisIO(GisLoader, GisParser, TupleMapper, BaseIO):
    pass


class WktIO(GisLoader, GisParser, WktMapper, BaseIO):
    pass



from collections import namedtuple, OrderedDict
import re


class BaseMapper(object):
    def get_key_field(self):
        return self.map_field(self.key_field)

    def map_field(self, field):
        return field

    def map_value(self, field, value):
        return value

    def unmap_field(self, field):
        return field

    def unmap_value(self, field, value):
        return value

    def usable_item(self, item):
        uitem = {}
        for key, val in item.iteritems():
            field = self.map_field(key)
            value = self.map_value(field, val)
            uitem[field] = value
        return uitem

    def parse_usable_item(self, uitem):
        item = {}
        for field, value in uitem.iteritems():
            key = self.unmap_field(field)
            val = self.unmap_value(field, value)
            item[key] = val
        return item


class DictMapper(BaseMapper):
    field_map = {}
    value_map = {}

    def map_field(self, field):
        field = self.field_map[field] if field in self.field_map else field
        return field

    def map_value(self, field, value):
        if not isinstance(value, basestring):
            return value
        value = self.value_map[value] if value in self.value_map else value
        return value

    def unmap_field(self, field):
        for f in self.field_map:
            if self.field_map[f] == field:
                return f
        return field

    def unmap_value(self, field, value):
        if not isinstance(value, basestring):
            return value
        for v in self.value_map:
            if self.value_map[v] == value:
                return v
        return value


class TupleMapper(DictMapper):
    @property
    def field_map(self):
        #FIXME: check for duplicates
        if not hasattr(self, '_field_map'):
            field_names = self.get_field_names()
            items = [
                (field, re.sub(r'\W', '', field.lower()))
                for field in field_names
            ]
            self._field_map = OrderedDict(items)
        return self._field_map

    @property
    def tuple_class(self):
        "Returns a class to use for individual items"

        if not hasattr(self, '_item_class'):
            cls = namedtuple(
                self.__class__.__name__ + 'Tuple',
                self.field_map.values()
            )
            self._tuple_class = cls

        return self._tuple_class

    @property
    def tuple_prototype(self):
        if not hasattr(self, '_tuple_prototype'):
            vals = {field: None for field in self.field_map.values()}
            self._tuple_prototype = self.tuple_class(**vals)
        return self._tuple_prototype

    def usable_item(self, item):
        mapped = super(TupleMapper, self).usable_item(item)
        return self.tuple_prototype._replace(**mapped)

    def parse_usable_item(self, uitem):
        mapped = {key: getattr(uitem, key) for key in self.field_map.values()}
        return super(TupleMapper, self).parse_usable_item(mapped)

    def create(self, **kwargs):
        return self.tuple_prototype._replace(**kwargs)



import json

try:
    import unicodecsv as csv
    UNICODE_CSV = True
except ImportError:
    import csv
    UNICODE_CSV = False

from lxml import etree as xml

from .base import BaseParser, TableParser


class CsvParser(TableParser):

    def parse(self):
        # Like DictReader, assume explicit field definition means CSV does not
        # contain column headers.
        fields = self.get_field_names()
        if self.start_row is None:
            if fields:
                self.start_row = 0
            else:
                self.start_row = 1
        if self.header_row is None:
            if fields:
                self.header_row = None
            else:
                self.header_row = 0

        self.csvdata = csv.DictReader(self.file, fields)
        self.field_names = self.csvdata.fieldnames
        self.data = [row for row in self.csvdata]

    def dump(self, file=None):
        if file is None:
            file = self.file
        args = file, self.get_field_names()
        kwargs = {'encoding': 'utf-8'} if UNICODE_CSV else {}
        csvout = csv.DictWriter(*args, **kwargs)
        csvout.writeheader()
        for row in self.data:
            csvout.writerow(row)


class JsonParser(BaseParser):
    indent = None
    namespace = None

    def parse(self):
        try:
            obj = json.load(self.file)
            if self.namespace:
                for key in self.namespace.split('.'):
                    obj = obj[key]
            self.data = map(self.parse_item, obj)
        except ValueError:
            self.data = []

    def parse_item(self, item):
        return item

    def dump(self, file=None):
        if file is None:
            file = self.file
        obj = map(self.dump_item, self.data)
        if self.namespace:
            for key in reversed(self.namespace.split('.')):
                obj = {key: obj}
        json.dump(obj, file, indent=self.indent)

    def dump_item(self, item):
        return item


class XmlParser(BaseParser):
    root_tag = None
    item_tag = None

    def parse(self):
        doc = xml.parse(self.file)
        root = doc.getroot()
        if self.root_tag is None:
            self.root_tag = root.tag
        if self.item_tag is None:
            self.item_tag = list(root)[0].tag
        self.data = map(self.parse_item, root.findall(self.item_tag))

    def parse_item(self, el):
        return {e.tag: e.text for e in el}

    def dump(self, file=None):
        if file is None:
            file = self.file
        root = xml.Element(self.root_tag)
        for item in self.data:
            root.append(self.dump_item(item))
        xml.ElementTree(root).write(file)

    def dump_item(self, item):
        el = xml.Element(self.item_tag)
        for key in self.get_field_names():
            if key not in item or item[key] is None:
                continue
            sel = xml.SubElement(el, key)
            sel.text = unicode(item.get(key))
        return el



from wq.io.base import BaseIO
from wq.io.loaders import FileLoader, BinaryFileLoader, NetLoader, StringLoader
from wq.io.parsers import CsvParser, JsonParser, XmlParser, ExcelParser
from wq.io.mappers import TupleMapper
import mimetypes

PARSERS = {
    'application/vnd.ms-excel': ExcelParser,
    'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet':
    ExcelParser,
    'text/csv': CsvParser,
    'application/json': JsonParser,
    'application/xml': XmlParser,
}
BINARY_FORMATS = (
    'application/vnd.ms-excel',
    'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
)


def make_io(loader, parser, mapper=TupleMapper, name=None):
    if name is None:
        lname = parser.__name__.replace('Parser', '')
        pname = loader.__name__.replace('Loader', '')
        if mapper == TupleMapper:
            mname = ""
        else:
            mname = mapper.__name__.replace('Mapper', '')
        name = lname + pname + mname + "IO"
    return type(name, (loader, parser, mapper, BaseIO), {})


def guess_type(filename):
    mimetype, encoding = mimetypes.guess_type(filename)
    if mimetype is None:
        try:
            import magic
            mimetype = magic.from_file(filename, mime=True)
        except ImportError:
            pass
    return mimetype


def load_file(filename, mapper=TupleMapper, options={}):
    mimetype = guess_type(filename)
    if mimetype not in PARSERS:
        raise Exception("Could not determine parser for %s" % mimetype)
    parser = PARSERS[mimetype]
    loader = BinaryFileLoader if mimetype in BINARY_FORMATS else FileLoader
    IO = make_io(loader, parser, mapper)
    return IO(filename=filename, **options)


def load_string(string, mapper=TupleMapper, options={}):
    if string.startswith('<'):
        parser = XmlParser
    elif string.startswith('[') or (
            string.startswith('{') and 'namespace' in options):
        parser = JsonParser
    elif ',' in string:
        parser = CsvParser
    else:
        raise Exception("Could not determine parser for string!")

    loader = StringLoader
    IO = make_io(loader, parser, mapper)
    return IO(string=string, **options)



from wq.io import XmlNetIO
from datetime import date, timedelta


class CocorahsIO(XmlNetIO):
    state = None
    county = None
    startdate = date.today() - timedelta(days=30)
    enddate = date.today()
    datetype = "reportdate"
    type = "Daily"

    url = "http://data.cocorahs.org/cocorahs/export/exportreports.aspx"
    params = {
        'dtf': "1",
        'Format': "XML",
        'TimesInGMT': "False",
        'responsefields': "all"
    }
    root_tag = 'Cocorahs'

    @property
    def item_tag(self):
        if self.type == "Daily":
            return 'DailyPrecipReports/DailyPrecipReport'
        elif self.type == "MultiDay":
            return 'MultiDayPrecipReports/MultiDayPrecipReport'
        raise Exception("%s is not a valid report type!" % self.type)

    def load(self):
        fmt = '%m/%d/%Y'
        self.params['ReportType'] = self.type
        self.params['State'] = self.state
        if self.county is not None:
            self.params['County'] = self.county

        self.params['ReportDateType'] = self.datetype
        if self.datetype == "reportdate":
            self.params['StartDate'] = self.startdate.strftime(fmt)
            self.params['EndDate'] = self.enddate.strftime(fmt)
        elif self.datetype == "timestamp":
            self.params['Date'] = self.startdate.strftime(fmt + " %I:%M %p")

        super(CocorahsIO, self).load()



from os.path import join, dirname
from setuptools import setup, find_packages

LONG_DESCRIPION = """
Consistent API for reading and writing to external datasets
"""

def long_description():
    """Return long description from README.rst if it's present
    because it doesn't get installed."""
    try:
        return open(join(dirname(__file__), 'README.rst')).read()
    except IOError:
        return LONG_DESCRIPTION

setup(
    name = 'wq.io',
    version = '0.2.0',
    author='S. Andrew Sheppard',
    author_email='andrew@wq.io',
    url='http://wq.io/wq.io',
    license='MIT',
    packages=find_packages(),
    namespace_packages=['wq'],
    description='Consistent iterable API for reading and writing to external datasets',
    long_description=long_description(),
    install_requires=['httplib2','lxml','xlrd'],
    classifiers = [
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
    ]
)



from httplib2 import Http
from StringIO import StringIO
from urllib import urlencode


class BaseLoader(object):
    def load(self):
        raise NotImplementedError


class FileLoader(BaseLoader):
    filename = None

    def load(self):
        try:
            self.file = open(self.filename)
        except:
            self.file = StringIO()

    def save(self):
        file = open(self.filename, 'w+')
        self.dump(file)
        file.close()
        self.load()


class BinaryFileLoader(FileLoader):
    def load(self):
        self.file = open(self.filename, 'rb')


class StringLoader(BaseLoader):
    string = ""

    def load(self):
        self.file = StringIO(self.string)

    def save(self):
        file = StringIO()
        self.dump(file)
        self.string = file.getvalue()
        file.close()
        self.load()


class NetLoader(BaseLoader):
    "NetLoader: opens HTTP/REST resources for use in wq.io"

    username = None
    password = None
    headers = {}

    http = Http()

    @property
    def url(self):
        raise NotImplementedError

    def load(self, **kwargs):

        if self.username is not None and self.password is not None:
            self.http.add_credentials(self.username, self.password)

        content = self.GET()
        self.file = StringIO(content)

    def req(self, url=None, method=None, params=None, body=None, headers={}):
        if url is None:
            url = self.url

        if params is None:
            params = getattr(self, 'params', None)

        if params is not None:
            url += '?' + urlencode(params)

        all_headers = self.headers.copy()
        all_headers.update(headers)

        resp, content = self.http.request(
            url, method=method, body=body, headers=all_headers
        )
        if resp.status < 200 or resp.status > 299:
            raise Exception(url + '\n' + content)

        return content

    def GET(self, **kwargs):
        return self.req(method='GET', **kwargs)

    def POST(self, **kwargs):
        return self.req(method='POST', **kwargs)

    def PUT(self, **kwargs):
        return self.req(method='PUT', **kwargs)

    def DELETE(self, **kwargs):
        return self.req(method='DELETE', **kwargs)



from wq.io.gis.gdal import DataSource, Layer, Feature
from wq.io.loaders import FileLoader
from wq.io.mappers import TupleMapper


class GisLoader(FileLoader):
    def load(self):
        self.datasource = DataSource(self.filename)

    def save(self):
        raise NotImplementedError


class GisParser(object):
    layer = None
    layer_name = None

    def parse(self):
        if len(self.datasource) == 1:
            self.layer_name = self.datasource[0].name

        if self.layer_name is None:
            self.data = [{'name': l.name, 'count': len(l), 'data': l}
                         for l in self.datasource]
            return

        for l in self.datasource:
            if l.name == self.layer_name:
                self.data = map(self.parse_feature, l)
                self.layer = l

    def parse_feature(self, f):
        feat = {key: f.get(key) for key in f.fields}
        feat['geometry'] = f.geom
        return feat

    def dump_feature(self, feat):
        raise NotImplementedError


class WktMapper(TupleMapper):
    def map_value(self, field, value):
        if field == 'geometry':
            return value.wkt
        else:
            return value
